@@ -1,4 +1,4 @@
-# $Id: CDDB.pm,v 1.24 2001/05/03 20:30:38 troc Exp troc $
+# $Id: CDDB.pm,v 1.25 2002/12/22 04:34:00 troc Exp $
 # Documentation and Copyright exist after __END__
 
 package CDDB;
@@ -8,7 +8,7 @@ use strict;
 use vars qw($VERSION);
 use Carp;
 
-$VERSION = '1.08';
+$VERSION = '1.16';
 
 BEGIN {
   if ($^O eq 'MSWin32') {
@@ -28,18 +28,21 @@ use Sys::Hostname;
 # list from freedb.freedb.org, which is a round-robin for all the
 # others anyway.
 
-my @cddbp_hosts =
-  ( [ 'localhost'         => 8880 ],
-    [ 'freedb.freedb.org' => 8880 ],
-    [ 'uk.freedb.org'     => 8880 ],
-    [ 'no.freedb.org'     => 8880 ],
-    [ 'de.freedb.org'     => 8880 ],
-    [ 'at.freedb.org'     => 8880 ],
-    [ 'freedb.freedb.de'  => 8880 ],
-  );
+my @cddbp_hosts = (
+  [ 'localhost'         => 8880 ],
+  [ 'freedb.freedb.org' => 8880 ],
+  [ 'us.freedb.org',    => 8880 ],
+  [ 'ca.freedb.org',    => 8880 ],
+  [ 'ca2.freedb.org',   => 8880 ],
+  [ 'uk.freedb.org'     => 8880 ],
+  [ 'no.freedb.org'     => 8880 ],
+  [ 'de.freedb.org'     => 8880 ],
+  [ 'at.freedb.org'     => 8880 ],
+  [ 'freedb.freedb.de'  => 8880 ],
+);
 
-# Determine whether we can submit changes by e-mail.  TODO: HTTP
-# submissions.
+#------------------------------------------------------------------------------
+# Determine whether we can submit changes by e-mail.
 
 my $imported_mail = 0;
 eval {
@@ -49,15 +52,25 @@ eval {
   $imported_mail = 1;
 };
 
-# Send a command.  If we're not connected, try to connect
-# first. Returns 1 if the command is sent ok; 0 if there was a
-# problem.
+#------------------------------------------------------------------------------
+# Determine whether we can use HTTP for requests and submissions.
+
+my $imported_http = 0;
+eval {
+  require LWP;
+  require HTTP::Request;
+  $imported_http = 1;
+};
+
+#------------------------------------------------------------------------------
+# Send a command.  If we're not connected, try to connect first.
+# Returns 1 if the command is sent ok; 0 if there was a problem.
 
 sub command {
   my $self = shift;
   my $str = join(' ', @_);
 
-  unless (exists $self->{handle}) {
+  unless ($self->{handle}) {
     $self->connect() or return 0;
   }
 
@@ -70,6 +83,7 @@ sub command {
   return 1;
 }
 
+#------------------------------------------------------------------------------
 # Retrieve a line from the server.  Uses a buffer to allow for
 # ungetting lines.  Returns the next line or undef if there is a
 # problem.
@@ -84,10 +98,10 @@ sub getline {
   }
 
   my $socket = $self->{handle};
-  return undef unless defined $socket;
+  return unless defined $socket;
 
   my $fd = fileno($socket);
-  return undef unless defined $fd;
+  return unless defined $fd;
 
   vec(my $rin = '', $fd, 1) = 1;
   my $timeout = $self->{timeout} || undef;
@@ -98,16 +112,17 @@ sub getline {
     # Fail if the socket is inactive for the timeout period.  Fail
     # also if sysread returns nothing.
 
-    return undef unless select(my $rout=$rin, undef, undef, $timeout);
-    return undef unless defined sysread($socket, my $buf='', 1024);
+    return unless select(my $rout=$rin, undef, undef, $timeout);
+    return unless defined sysread($socket, my $buf='', 1024);
 
     $frame .= $buf;
     my @lines = split(/\x0D?\x0A/, $frame);
-    $frame = ( (length($buf) == 0 || substr($buf, -1, 1) eq "\x0A")
-               ? ''
-               : pop(@lines)
-             );
-    push @{$self->{lines}}, @lines;
+    $frame = (
+      (length($buf) == 0 || substr($buf, -1, 1) eq "\x0A")
+      ? ''
+      : pop(@lines)
+    );
+    push @{$self->{lines}}, map { decode('utf8', $_) } @lines;
   }
 
   $self->{frame} = $frame;
@@ -117,6 +132,7 @@ sub getline {
   return $line;
 }
 
+#------------------------------------------------------------------------------
 # Receive a server response, and parse it into its numeric code and
 # text message.  Return the code's first character, which usually
 # indicates the response class (ok, error, information, warning,
@@ -128,16 +144,17 @@ sub response {
 
   my $str = $self->getline();
 
-  return undef unless defined($str);
+  return unless defined($str);
 
   # Fail if the line we get isn't the proper format.
-  return undef unless ( ($code, $text) = ($str =~ /^(\d+)\s*(.*?)\s*$/) );
+  return unless ( ($code, $text) = ($str =~ /^(\d+)\s*(.*?)\s*$/) );
 
   $self->{response_code} = $code;
   $self->{response_text} = $text;
   substr($code, 0, 1);
 }
 
+#------------------------------------------------------------------------------
 # Accessors to retrieve the last response() call's code and text
 # separately.
 
@@ -151,6 +168,7 @@ sub text {
   $self->{response_text};
 }
 
+#------------------------------------------------------------------------------
 # Helper to print stuff for debugging.
 
 sub debug_print {
@@ -164,6 +182,7 @@ sub debug_print {
   print STDERR $text, "\n";
 }
 
+#------------------------------------------------------------------------------
 # Read data until it's terminated by a single dot on its own line.
 # Two dots at the start of a line are replaced by one.  Returns an
 # ARRAY reference containing the lines received, or undef on error.
@@ -173,7 +192,7 @@ sub read_until_dot {
   my @lines;
 
   while ('true') {
-    my $line = $self->getline() or return undef;
+    my $line = $self->getline() or return;
     last if ($line =~ /^\.$/);
     $line =~ s/^\.\././;
     push @lines, $line;
@@ -182,6 +201,7 @@ sub read_until_dot {
   \@lines;
 }
 
+#------------------------------------------------------------------------------
 # Create an object to represent one or more cddbp sessions.
 
 sub new {
@@ -195,17 +215,19 @@ sub new {
   my $login = $param{Login} || $ENV{LOGNAME} || $ENV{USER};
   if (not defined $login) {
     if (USING_WINDOWS) {
-      carp( "Can't get login ID.  Use Login parameter or " .
-            "set LOGNAME or USER environment variable.  Using default login " .
-            "ID 'win32usr'"
-          );
+      carp(
+        "Can't get login ID.  Use Login parameter or " .
+        "set LOGNAME or USER environment variable.  Using default login " .
+        "ID 'win32usr'"
+      );
       $login = 'win32usr';
     }
     else {
       $login = getpwuid($>)
-        or croak( "Can't get login ID.  " .
-                  "Set LOGNAME or USER environment variable and try again: $!"
-                );
+        or croak(
+          "Can't get login ID.  " .
+          "Set LOGNAME or USER environment variable and try again: $!"
+        );
     }
   }
 
@@ -219,7 +241,7 @@ sub new {
 
   # Choose a particular cddbp port.
   my $port = $param{Port};
-  $port = 0 unless defined $port;
+  $port = 8880 unless $port;
 
   # Choose a particular cddbp submission address.
   my $submit_to = $param{Submit_Address};
@@ -233,15 +255,35 @@ sub new {
   my $client_version = $param{Client_Version};
   $client_version = $VERSION unless defined $client_version;
 
+  # Whether to use utf-8 for submission
+  my $utf8 = $param{Utf8};
+  $utf8 = 1 unless defined $utf8;
+  if ($utf8) {
+    eval {
+      require Encode;
+      import Encode;
+    };
+    if ( $@ ) {
+      carp 'Unable to load the Encode module, falling back to ascii';
+      $utf8 = 0;
+    }
+  }
+
+  eval 'sub encode { $_[1] };sub decode { $_[1] }' unless $utf8;
+
   # Change the cddbp protocol level.
   my $cddb_protocol = $param{Protocol_Version};
-  $cddb_protocol = 1 unless defined $cddb_protocol;
+  $cddb_protocol = ($utf8 ? 6 : 1) unless defined $cddb_protocol;
+  carp <<EOF if $utf8 and $cddb_protocol < 6;
+You have requested protocol level $cddb_protocol. However,
+utf-8 support is only available starting from level 6
+EOF
 
   # Mac Freaks Got Spaces!  Augh!
   $login =~ s/\s+/_/g;
 
-  my $self = bless
-  { hostname      => $hostname,
+  my $self = bless {
+    hostname      => $hostname,
     login         => $login,
     mail_from     => undef,
     mail_host     => undef,
@@ -252,6 +294,7 @@ sub new {
     host          => $host,
     port          => $port,
     cddb_protocol => $cddb_protocol,
+    utf8          => $utf8,
     lines         => [],
     frame         => '',
     response_code => '000',
@@ -261,12 +304,13 @@ sub new {
   $self;
 }
 
+#------------------------------------------------------------------------------
 # Disconnect from a cddbp server.  This is needed sometimes when a
 # server decides a session has performed enough requests.
 
 sub disconnect {
   my $self = shift;
-  if (exists $self->{handle}) {
+  if ($self->{handle}) {
     $self->command('quit');     # quit
     $self->response();          # wait for any response
     delete $self->{handle};     # close the socket
@@ -276,6 +320,7 @@ sub disconnect {
   }
 }
 
+#------------------------------------------------------------------------------
 # Connect to a cddbp server.  Connecting and disconnecting are done
 # transparently and are performed on the basis of need.  Furthermore,
 # this routine will cycle through servers until one connects or it has
@@ -317,7 +362,9 @@ HOST:
         # None of the servers worked.  Time to leave.
         unless (@cddbp_hosts) {
           $self->debug_print( 0, "--- all cddbp servers failed to answer" );
-          return undef;
+          warn "No cddb protocol servers answer.  Is your network OK?\n"
+            unless $self->{debug};
+          return;
         }
 
         $cddbp_host = shift(@cddbp_hosts);
@@ -325,24 +372,24 @@ HOST:
       }
 
       # Assign the host we selected, and attempt a connection.
-      $self->debug_print
-        ( 0,
-          "=== connecting to $self->{host} port $self->{port}"
-        );
-      $self->{handle} = new IO::Socket::INET
-        ( PeerAddr => $self->{host},
-          PeerPort => $self->{port},
-          Proto    => 'tcp',
-          Timeout  => 30,
-        );
+      $self->debug_print(
+        0,
+        "=== connecting to $self->{host} port $self->{port}"
+      );
+      $self->{handle} = new IO::Socket::INET(
+        PeerAddr => $self->{host},
+        PeerPort => $self->{port},
+        Proto    => 'tcp',
+        Timeout  => 30,
+      );
 
       # The host did not answer.  Clean up after the failed attempt
       # and cycle to the next host.
       unless (defined $self->{handle}) {
-        $self->debug_print
-          ( 0,
-            "--- error connecting to $self->{host} port $self->{port}: $!"
-          );
+        $self->debug_print(
+          0,
+          "--- error connecting to $self->{host} port $self->{port}: $!"
+        );
         delete $self->{handle};
         $self->{host} = $self->{port} = '';
         next HOST;
@@ -351,10 +398,10 @@ HOST:
       # The host accepted our connection.  We'll push it back on the
       # list of known cddbp hosts so it can be tried later.  And we're
       # done with the host list cycle for now.
-      $self->debug_print
-        ( 0,
-          "+++ successfully connected to $self->{host} port $self->{port}"
-        );
+      $self->debug_print(
+        0,
+        "+++ successfully connected to $self->{host} port $self->{port}"
+      );
       push(@cddbp_hosts, $cddbp_host);
       last HOST;
     }
@@ -368,28 +415,32 @@ HOST:
     # Get the server's banner message.  Try reconnecting if it's bad.
     my $code = $self->response();
     if ($code != 2) {
-      $self->debug_print( 0, "--- bad cddbp response: ",
-                          $self->code(), ' ', $self->text()
-                        );
+      $self->debug_print(
+        0, "--- bad cddbp response: ",
+        $self->code(), ' ', $self->text()
+      );
       next HANDSHAKE;
     }
 
     # Say hello, and wait for a response.
-    $self->command( 'cddb hello',
-                    $self->{login}, $self->{hostname},
-                    $self->{libname}, $self->{libver}
-                  );
+    $self->command(
+      'cddb hello',
+       $self->{login}, $self->{hostname},
+       $self->{libname}, $self->{libver}
+    );
     $code = $self->response();
     if ($code == 4) {
-      $self->debug_print( 0, "--- the server denies us: ",
-                          $self->code(), ' ', $self->text()
-                        );
-      return undef;
+      $self->debug_print(
+        0, "--- the server denies us: ",
+        $self->code(), ' ', $self->text()
+      );
+      return;
     }
     if ($code != 2) {
-      $self->debug_print( 0, "--- the server didn't handshake: ",
-                          $self->code(), ' ', $self->text()
-                        );
+      $self->debug_print(
+        0, "--- the server didn't handshake: ",
+        $self->code(), ' ', $self->text()
+      );
       next HANDSHAKE;
     }
 
@@ -398,11 +449,12 @@ HOST:
       $self->command( 'proto', $self->{cddb_protocol} );
       $code = $self->response();
       if ($code != 2) {
-        $self->debug_print( 0, "--- can't set protocol level ",
-                            $self->{cddb_protocol}, ' ',
-                            $self->code(), ' ', $self->text()
-                          );
-        return undef;
+        $self->debug_print(
+          0, "--- can't set protocol level ",
+          $self->{cddb_protocol}, ' ',
+          $self->code(), ' ', $self->text()
+        );
+        return;
       }
     }
 
@@ -418,9 +470,10 @@ sub DESTROY {
   $self->disconnect();
 }
 
-#------------------------------------------------------------------------------
+###############################################################################
 # High-level cddbp functions.
 
+#------------------------------------------------------------------------------
 # Get a list of available genres.  Returns an array of genre names, or
 # undef on failure.
 
@@ -430,20 +483,22 @@ sub get_genres {
 
   $self->command('cddb lscat');
   my $code = $self->response();
-  return undef unless $code;
+  return unless $code;
 
   if ($code == 2) {
     my $genres = $self->read_until_dot();
     return @$genres if defined $genres;
-    return undef;
+    return;
   }
 
-  $self->debug_print( 0, '--- error listing categories: ',
-                      $self->code(), ' ', $self->text()
-                    );
-  return undef;
+  $self->debug_print(
+    0, '--- error listing categories: ',
+    $self->code(), ' ', $self->text()
+  );
+  return;
 }
 
+#------------------------------------------------------------------------------
 # Calculate a cddbp ID based on a text table of contents.  The text
 # format was chosen because it was straightforward and easy to
 # generate.  In a scalar context, this returns just the cddbp ID.  In
@@ -456,19 +511,22 @@ sub calculate_id {
   my $self = shift;
   my @toc = @_;
 
-  my ($seconds_previous, $seconds_first, $seconds_last, $cddbp_sum,
-      @track_numbers, @track_lengths, @track_offsets,
-     );
+  my (
+    $seconds_previous, $seconds_first, $seconds_last, $cddbp_sum,
+    @track_numbers, @track_lengths, @track_offsets,
+  );
 
   foreach my $line (@toc) {
     my ($track, $mm_begin, $ss_begin, $ff_begin) = split(/\s+/, $line, 4);
-    my $seconds_begin = ($mm_begin * 60) + $ss_begin;
+    my $frame_offset = (($mm_begin * 60 + $ss_begin) * 75) + $ff_begin + 150;
+    my $seconds_begin = int($frame_offset / 75);
 
     if (defined $seconds_previous) {
       my $elapsed = $seconds_begin - $seconds_previous;
-      push( @track_lengths,
-            sprintf("%02d:%02d", int($elapsed / 60), $elapsed % 60)
-          );
+      push(
+        @track_lengths,
+        sprintf("%02d:%02d", int($elapsed / 60), $elapsed % 60)
+      );
     }
     else {
       $seconds_first = $seconds_begin;
@@ -483,35 +541,54 @@ sub calculate_id {
     # Track 1000 was chosen for error information.
     if ($track == 1000) {
       $self->debug_print( 0, "error in TOC: $ff_begin" );
-      return undef;
+      return;
     }
 
     map { $cddbp_sum += $_; } split(//, $seconds_begin);
-    push @track_offsets, ($mm_begin * 60 + $ss_begin) * 75 + $ff_begin;
+    push @track_offsets, $frame_offset;
     push @track_numbers, sprintf("%03d", $track);
     $seconds_previous = $seconds_begin;
   }
 
   # Calculate the ID.  Whee!
-  my $total_seconds = $seconds_last - $seconds_first;
-  my $id = sprintf
-    ( "%08x",
-      (($cddbp_sum % 255) << 24)
-      | ($total_seconds << 8)
-      | scalar(@track_offsets)
-    );
+  my $id = sprintf(
+    "%08x",
+    (($cddbp_sum % 255) << 24)
+    | (($seconds_last - $seconds_first) << 8)
+    | scalar(@track_offsets)
+  );
 
   # In list context, we return several things.  Some of them are
   # useful for generating filenames or playlists (the padded track
   # numbers).  Others are needed for cddbp queries.
-  return
-    ($id, \@track_numbers, \@track_lengths, \@track_offsets, $total_seconds)
-      if wantarray();
+  return (
+    $id, \@track_numbers, \@track_lengths, \@track_offsets, $seconds_last
+  ) if wantarray();
 
   # Just return the cddbp ID in scalar context.
   return $id;
 }
 
+#------------------------------------------------------------------------------
+# Parse cdinfo's output so calculate_id() can eat it.
+
+sub parse_cdinfo {
+  my ($self, $command) = @_;
+  open(FH, $command) or croak "could not open `$command': $!";
+
+  my @toc;
+  while (<FH>) {
+    if (/(\d+):\s+(\d+):(\d+):(\d+)/) {
+      my @track = ($1,$2,$3,$4);
+      $track[0] = 999 if /leadout/;
+      push @toc, "@track";
+    }
+  }
+  close FH;
+  return @toc;
+}
+
+#------------------------------------------------------------------------------
 # Get a list of discs that match a particular CD's table of contents.
 # This accepts the TOC information as returned by calculate_id().  It
 # will also accept information in mp3 format, but I forget what that
@@ -543,38 +620,38 @@ ATTEMPT:
   while ('true') {
 
     # Send a cddbp query command.
-    $self->command( 'cddb query', $id, $track_count,
-                    $offsets_string, $total_seconds
-                  ) or return undef;
+    $self->command(
+      'cddb query', $id, $track_count,
+      $offsets_string, $total_seconds
+    ) or return;
 
     # Get the response.  Try again if the server is temporarly
     # unavailable.
     $code = $self->response();
-    if ($self->code() == 417) {
-      next ATTEMPT;
-    }
+    next ATTEMPT if $self->code() == 417;
     last ATTEMPT;
   }
 
   # Return undef if there's a problem.
-  return undef unless defined $code and $code == 2;
+  return unless defined $code and $code == 2;
 
   # Single matching disc.
   if ($self->code() == 200) {
-    my ($genre, $cddbp_id, $title) =
-      ($self->text() =~ /^(\S+)\s*(\S+)\s*(.*?)\s*$/);
+    my ($genre, $cddbp_id, $title) = (
+      $self->text() =~ /^(\S+)\s*(\S+)\s*(.*?)\s*$/
+    );
     return [ $genre, $cddbp_id, $title ];
   }
 
   # No matching discs.
-  return () if $self->code() == 202;
+  return if $self->code() == 202;
 
   # Multiple matching discs.
   # 210 Found exact matches, list follows (...)   [proto>=4]
   # 211 Found inexact matches, list follows (...) [proto>=1]
   if ($self->code() == 210 or $self->code() == 211) {
     my $discs = $self->read_until_dot();
-    return undef unless defined $discs;
+    return unless defined $discs;
 
     my @matches;
     foreach my $disc (@$discs) {
@@ -586,12 +663,14 @@ ATTEMPT:
   }
 
   # What the heck?
-  $self->debug_print(0, "--- unknown cddbp response: ",
-                     $self->code(), ' ', $self->text()
-                    );
-  return undef;
+  $self->debug_print(
+    0, "--- unknown cddbp response: ",
+    $self->code(), ' ', $self->text()
+  );
+  return;
 }
 
+#------------------------------------------------------------------------------
 # A little helper to combine list-context calculate_id() with
 # get_discs().
 
@@ -604,6 +683,19 @@ sub get_discs_by_toc {
   @discs;
 }
 
+#------------------------------------------------------------------------------
+# A little helper to get discs from an existing query string.
+# Contributed by Ron Grabowski.
+
+sub get_discs_by_query {
+  my ($self, $query) = @_;
+  my (undef, undef, $cddbp_id, $tracks, @offsets) = split /\s+/, $query;
+  my $total_seconds = pop @offsets;
+  my @discs = $self->get_discs($cddbp_id, \@offsets, $total_seconds);
+  return @discs;
+}
+
+#------------------------------------------------------------------------------
 # Retrieve the database record for a particular genre/id combination.
 # Returns a moderately complex hashref representing the cddbp record,
 # or undef on failure.
@@ -616,33 +708,41 @@ sub get_disc_details {
   # force a disconnect/reconnect here if we already did one.
   if (exists $self->{'got tracks before'}) {
     $self->disconnect();
-    $self->connect() or return undef;
+    $self->connect() or return;
   }
   $self->{'got tracks before'} = 'yes';
 
   $self->command('cddb read', $genre, $id);
   my $code = $self->response();
   if ($code != 2) {
-    $self->debug_print( 0, "--- cddbp host could not read the disc record: ",
-                        $self->code(), ' ', $self->text()
-                      );
-    return undef;
+    $self->debug_print(
+      0, "--- cddbp host could not read the disc record: ",
+      $self->code(), ' ', $self->text()
+    );
+    return;
   }
 
   my $track_file;
   unless (defined($track_file = $self->read_until_dot())) {
     $self->debug_print( 0, "--- cddbp disc record interrupted" );
-    return undef;
+    return;
   }
 
   # Parse that puppy.
+  return parse_xmcd_file($track_file, $genre);
+}
 
-  my @track_file = @$track_file;
-  my %details = ( offsets => [ ],
-                  seconds => [ ],
-                );
+# Arf!
+
+sub parse_xmcd_file {
+  my ($track_file, $genre) = @_;
+
+  my %details = (
+    offsets => [ ],
+    seconds => [ ],
+  );
   my $state = 'beginning';
-  foreach my $line (@track_file) {
+  foreach my $line (@$track_file) {
     # Keep returned so-called xmcd record...
     $details{xmcd_record} .= $line . "\n";
 
@@ -659,9 +759,10 @@ sub get_disc_details {
         next;
       }
       $state = 'headers';
-      next;
+      # This passes through on purpose.
     }
 
+    # This is not an elsif on purpose.
     if ($state eq 'headers') {
       if ($line =~ /^\#/) {
         $line =~ s/\s+/ /g;
@@ -676,9 +777,9 @@ sub get_disc_details {
 
     # This is not an elsif on purpose.
     if ($state eq 'data') {
-      next unless (my ($tag, $idx, $val) =
-                   ($line =~ /^\s*(.+?)(\d*)\s*\=\s*(.+?)\s*$/)
-                  );
+      next unless (
+        my ($tag, $idx, $val) = ($line =~ /^\s*(.+?)(\d*)\s*\=\s*(.+?)\s*$/)
+      );
       $tag = lc($tag);
 
       if ($idx ne '') {
@@ -705,9 +806,7 @@ sub get_disc_details {
 
   my $last_offset = 0;
   foreach (@{$details{offsets}}) {
-    push( @{$details{seconds}},
-          int(($_ - $last_offset) / 75)
-        );
+    push @{$details{seconds}}, int(($_ - $last_offset) / 75);
     $last_offset = $_;
   }
 
@@ -718,16 +817,21 @@ sub get_disc_details {
   $disc_length =~ s/ .*$//;
 
   my $first_start = shift @{$details{seconds}};
-  push( @{$details{seconds}},
-        $disc_length - int($details{offsets}->[-1] / 75) + 1 - $first_start
-      );
+  push(
+    @{$details{seconds}},
+    $disc_length - int($details{offsets}->[-1] / 75) + 1 - $first_start
+  );
 
-  \%details;
+  # Add the genre, if we have it.
+  $details{genre} = $genre;
+
+  return \%details;
 }
 
-#------------------------------------------------------------------------------
+###############################################################################
 # Evil voodoo e-mail submission stuff.
 
+#------------------------------------------------------------------------------
 # Return true/false whether the libraries needed to submit discs are
 # present.
 
@@ -736,6 +840,7 @@ sub can_submit_disc {
   $imported_mail;
 }
 
+#------------------------------------------------------------------------------
 # Build an e-mail address, and return it.  Caches the last built
 # address, and returns that on subsequent calls.
 
@@ -745,6 +850,7 @@ sub get_mail_address {
   return $self->{mail_from} = $self->{login} . '@' . $self->{hostname};
 }
 
+#------------------------------------------------------------------------------
 # Build an e-mail host, and return it.  Caches the last built e-mail
 # host, and returns that on subsequent calls.
 
@@ -771,17 +877,16 @@ sub submit_disc {
   my $self = shift;
   my %params = @_;
 
-  croak( "submit_disc needs " .
-         "Mail::Internet, Mail::Header, and MIME::QuotedPrint"
-       )
-    unless $imported_mail;
+  croak(
+    "submit_disc needs Mail::Internet, Mail::Header, and MIME::QuotedPrint"
+  ) unless $imported_mail;
 
   # Try yet again to fetch the hostname.  Fail if we cannot.
   unless (defined $self->{hostname}) {
     $self->{hostname} = &hostname() or croak "can't get hostname: $!";
   }
 
-  # Validate the required submission fields.
+  # Validate the required submission fields.  XXX Duplicated code.
   (exists $params{Genre})       or croak "submit_disc needs a Genre";
   (exists $params{Id})          or croak "submit_disc needs an Id";
   (exists $params{Artist})      or croak "submit_disc needs an Artist";
@@ -789,6 +894,19 @@ sub submit_disc {
   (exists $params{TrackTitles}) or croak "submit_disc needs TrackTitles";
   (exists $params{Offsets})     or croak "submit_disc needs Offsets";
   (exists $params{Revision})    or croak "submit_disc needs a Revision";
+  if (exists $params{Year}) {
+    unless ($params{Year} =~ /^\d{4}$/) {
+      croak "submit_disc needs a 4 digit year";
+    }
+  }
+  if (exists $params{GenreLong}) {
+    unless ($params{GenreLong} =~ /^([A-Z][a-zA-Z0-9]*\s?)+$/) {
+      croak(
+        "GenreLong must start with a capital letter and contain only " .
+        "letters and numbers"
+      );
+    }
+  }
 
   # Try to find a mail host.  We could probably grab the MX record for
   # the current machine, but that would require yet more strange
@@ -806,28 +924,39 @@ sub submit_disc {
   # Build the submission's headers.
   my $header = new Mail::Header;
   $header->add( 'MIME-Version' => '1.0' );
-  $header->add( 'Content-Type' => 'text/plain; charset=iso-8859-1' );
+  my $charset = $self->{'utf8'} ? 'utf-8' : 'iso-8859-1';
+  $header->add( 'Content-Type' => "text/plain; charset=$charset" );
   $header->add( 'Content-Disposition' => 'inline' );
   $header->add( 'Content-Transfer-Encoding' => 'quoted-printable' );
   $header->add( From    => $from );
   $header->add( To      => $self->{cddbmail} );
+  # send a copy to ourselves if we are debugging
+  $header->add( Cc => $from ) if $self->{debug};
   $header->add( Subject => "cddb $params{Genre} $params{Id}" );
 
   # Build the submission's body.
-  my @message_body =
-    ( '# xmcd',
-      '#',
-      '# Track frame offsets:',
-      map({ "#\t" . $_; } @{$params{Offsets}}),
-      '#',
-      '# Disc length: ' . (hex(substr($params{Id},2,4))+2) . ' seconds',
-      '#',
-      "# Revision: " . $params{Revision},
-      '# Submitted via: ' . $self->{libname} . ' ' . $self->{libver},
-      '#',
-      'DISCID=' . $params{Id},
-      'DTITLE=' . $params{Artist} . ' / ' . $params{DiscTitle},
-    );
+  my @message_body = (
+    '# xmcd',
+    '#',
+    '# Track frame offsets:',
+    map({ "#\t" . $_; } @{$params{Offsets}}),
+    '#',
+    '# Disc length: ' . (hex(substr($params{Id},2,4))+2) . ' seconds',
+    '#',
+    "# Revision: " . $params{Revision},
+    '# Submitted via: ' . $self->{libname} . ' ' . $self->{libver},
+    '#',
+    'DISCID=' . $params{Id},
+    'DTITLE=' . $params{Artist} . ' / ' . $params{DiscTitle},
+  );
+
+  # add year and genre
+  if (exists $params{Year}) {
+    push @message_body, 'DYEAR='.$params{Year};
+  }
+  if (exists $params{GenreLong}) {
+    push @message_body, 'DGENRE='.$params{GenreLong};
+  }
 
   # Dump the track titles.
   my $number = 0;
@@ -850,15 +979,15 @@ sub submit_disc {
   # The cddbp submissions daemon will barf if it's not.
   foreach my $line (@message_body) {
     $line .= "\n";
-    $line = MIME::QuotedPrint::encode_qp($line);
+    $line = MIME::QuotedPrint::encode_qp(encode('utf8', $line));
   }
 
   # Bundle the headers and body into an Internet mail.
-  my $mail = new Mail::Internet
-    ( undef,
-      Header => $header,
-      Body   => \@message_body,
-    );
+  my $mail = new Mail::Internet(
+    undef,
+    Header => $header,
+    Body   => \@message_body,
+  );
 
   # Try to send it using the "mail" utility.  This is commented out:
   # it strips the MIME headers from the message, invalidating the
@@ -883,7 +1012,7 @@ sub submit_disc {
 
   # Augh!  Everything failed!
   $self->debug_print( 0, '--- could not find a way to submit a disc' );
-  return undef;
+  return;
 }
 
 ###############################################################################
@@ -993,9 +1122,16 @@ B<Host> and B<Port> describe the cddbp server to connect to.  These
 default to 'freedb.freedb.org' and 8880, which is a multiplexor for
 all the other freedb servers.
 
+B<Utf8> is a boolean flag. If true, utf-8 will be used when submitting 
+CD info, and for interpreting the data reveived. This requires the 
+L<Encode> module (and probably perl version at least 5.8.0). The 
+default is true if the L<Encode> module can be loaded. Otherwise, it 
+will be false, meaning we fall back to ASCII.
+
 B<Protocol_Version> sets the cddbp version to use.  CDDB.pm will not
 connect to servers that don't support the version specified here.  The
-requested protocol version defaults to 1 if omitted.
+requested protocol version defaults to 1 if B<Utf8> is off, and to 
+6 if it is on.
 
 B<Login> is the login ID you want to advertise to the cddbp server.
 It defaults to the login ID your computer assigns you, if that can be
@@ -1005,7 +1141,8 @@ On Windows systems, it defaults to "win32usr" if no default method can
 be found and no Login parameter is set.
 
 B<Submit_Address> is the e-mail address where new disc submissions go.
-This defaults to 'freedb-submit@freedb.org'.
+This defaults to 'freedb-submit@freedb.org'. Note, that testing 
+submissions should be done via C<test-submit@freedb.org>.
 
 B<Client_Name> and B<Client_Version> describe the client software used
 to connect to the cddbp server.  They default to 'CDDB.pm' and
@@ -1016,6 +1153,8 @@ B<Debug> enables verbose operational information on STDERR when set to
 true.  It's normally not needed, but it can help explain why a program
 is failing.  If someone finds a reproduceable bug, the Debug output
 and a test program would be a big help towards having it fixed.
+In case of submission, if this flag is on, a copy of the submission 
+e-mail will be sent to the I<From> address.
 
 =item get_genres
 
@@ -1038,6 +1177,9 @@ FreeBSD.  CDDB.pm's author has used external programs like dagrab to
 fetch the offsets.  Actual CDs aren't always necessary: the author has
 heard of people generating TOC information from mp3 file lengths.
 
+That said, see parse_cdinfo() for a routine to parse "cdinfo" output
+into a table of contents list suitable for calculate_id().
+
 calculate_id() accepts TOC information as a list of strings.  Each
 string contains four fields, separated by whitespace:
 
@@ -1128,9 +1270,13 @@ seconds.  The get_discs() function needs it.
 =item get_discs CDDBP_ID, TRACK_OFFSETS, TOTAL_SECONDS
 
 get_discs() asks the cddbp server for a summary of all the CDs
-matching a given cddbp ID, track offsets, and total playing time.  It
-accepts these in the same format returned by calculate_id(), so it's
-easy to pass the values from one function to the other.
+matching a given cddbp ID, track offsets, and total playing time.
+These values can be retrieved from calculade_id().
+
+  my @id_info       = $cddbp->calculate_id(@toc);
+  my $cddbp_id      = $id_info->[0];
+  my $track_offsets = $id_info->[3];
+  my $total_seconds = $id_info->[4];
 
 get_discs() returns an array of matching discs, each of which is
 represented by an array reference.  It returns an empty array if the
@@ -1164,6 +1310,16 @@ This function acts as a macro, combining calculate_id() and
 get_discs() calls into one function.  It takes the same parameters as
 calculate_id(), and it returns the same information as get_discs().
 
+=item get_discs_by_query QUERY_STRING
+
+Fetch discs by a pre-built cddbp query string.  Some disc querying
+programs report this string, and get_discs_by_query() is a convenient
+way to use that.
+
+Cddb protocol query strings look like:
+
+  cddb query $cddbp_id $track_count @offsets $total_seconds
+
 =item get_disc_details DISC_GENRE, CDDBP_ID
 
 This function fetches a disc's detailed information from a cddbp
@@ -1230,6 +1386,17 @@ $disc_details->{xmcd_record}
 The xmcd_record field contains a copy of the entire unprocessed cddbp
 response that generated all the other fields.
 
+$disc_details->{genre}
+
+This is merely a copy of DISC_GENRE, since it's otherwise not possible
+to determine it from the hash.
+
+=item parse_xmcd_file XMCD_FILE_CONTENTS, [GENRE]
+
+Parses an array ref of lines read from an XMCD file into the
+disc_details hash described above.  If the GENRE parameter is set it
+will be included in disc_details.
+
 =item can_submit_disc
 
 Returns true or false, depending on whether CDDB.pm has enough
@@ -1272,6 +1439,17 @@ can see it and have an opportunity to correct it.
                        Host => $mail_host,
                      );
 
+=item parse_cdinfo CDINFO_FILE
+
+Generates a table of contents suitable for calculate_id() based on the
+output of a program called "cdinfo".  CDINFO_FILE may either be a text
+file, or it may be the cdinfo program itself.
+
+  my @toc = parse_cdinfo("cdinfo.txt"); # read cdinfo.txt
+  my @toc = parse_cdinfo("cdinfo|");    # run cdinfo directly
+
+The table of contents can be passed directly to calculate_id().
+
 =item submit_disc DISC_DETAILS
 
 submit_disc() submits a disc record to a cddbp server.  Currently it
@@ -1341,6 +1519,13 @@ required, and CDDB.pm will try to figure one out on its own.  It will
 look at the SMTPHOSTS environment variable is not defined, it will try
 'mail' and 'localhost' before finally failing.
 
+Revision => REVISION
+
+The revision number. Should be 1 for new submissions, and one higher 
+than the previous one for updates. The previous revision number is 
+available as the C<revision> field in the hash returned by 
+get_disc_details().
+
 =back
 
 =head1 PRIVATE METHODS
@@ -1359,8 +1544,12 @@ done.
 
 =head1 CONTACT AND COPYRIGHT
 
-Copyright 1998-2002 Rocco Caputo E<lt>troc+cddb@pobox.com<gt>.  All
-rights reserved.  This program is free software; you can redistribute
-it and/or modify it under the same terms as Perl itself.
+Copyright 1998-2002 Rocco Caputo.  All rights reserved.  This program
+is free software; you can redistribute it and/or modify it under the
+same terms as Perl itself.
+
+Rocco may be contacted at rcaputo@cpan.org.
 
 =cut
+
+# vim: sw=2 tw=70:
